
#include "TComUpsampleFilter.h"
#include "TypeDef.h"

#if SVC_EXTENSION
// ====================================================================================================================
// Tables
// ====================================================================================================================

const Int TComUpsampleFilter::m_iFilter16_luma_fixed[12][NTAPS_US_LUMA] =
{
  {0,  0,   0, 64,  0,   0, 0,  0},
  {-1, 1,  -4, 63,  6,  -2, 1,  0},
  {-1, 3,  -7, 61, 12,  -5, 2, -1},
  {-1, 4, -10, 57, 19,  -7, 3, -1}, 
  {-1, 4, -11, 52, 26,  -8, 3, -1},
  {-1, 5, -12, 47, 32,  -9, 3, -1},      
  {-1, 4, -11, 40, 40, -11, 4, -1},      
  {-1, 3,  -9, 32, 47, -12, 5, -1},
  {-1, 3,  -8, 26, 52, -11, 4, -1},
  {-1, 3,  -7, 19, 57, -10, 4, -1},
  {-1, 2,  -5, 12, 61,  -7, 3, -1},
  { 0, 1,  -2,  6, 63,  -4, 1, -1} 
};

const Int TComUpsampleFilter::m_iFilter16_chroma_fixed[12][NTAPS_US_CHROMA] =
{
  { 0, 64,  0,  0},
  {-2, 62,  5, -1},
  {-4, 59, 11, -2},
  {-4, 54, 16, -2},
  {-5, 50, 22, -3},
  {-5, 43, 30, -4},
  {-4, 36, 36, -4},
  {-4, 30, 43, -5},
  {-3, 22, 50, -5},
  {-2, 16, 54, -4},
  {-2, 11, 59, -4},
  {-1,  5, 62, -2}
};

TComUpsampleFilter::TComUpsampleFilter(void)
{
}

TComUpsampleFilter::~TComUpsampleFilter(void)
{
}


#define sum_luma_hor(pel, coeff) ( pel[0]*coeff[0] + pel[1]*coeff[1] + pel[2]*coeff[2] + pel[3]*coeff[3] + pel[4]*coeff[4] + pel[5]*coeff[5] + pel[6]*coeff[6] + pel[7]*coeff[7])
#define sum_chroma_hor(pel, coeff) ( pel[0]*coeff[0] + pel[1]*coeff[1] + pel[2]*coeff[2] + pel[3]*coeff[3] )

#define sum_luma_ver(pel, coeff, stride) ( pel[0]*coeff[0] + pel[stride]*coeff[1] + pel[2*stride]*coeff[2] + pel[3*stride]*coeff[3] + pel[4*stride]*coeff[4] + pel[5*stride]*coeff[5] + pel[6*stride]*coeff[6] + pel[7*stride]*coeff[7])
#define sum_chroma_ver(pel, coeff, stride) ( pel[0]*coeff[0] + pel[stride]*coeff[1] + pel[2*stride]*coeff[2] + pel[3*stride]*coeff[3] )

Void TComUpsampleFilter::upsampleBasePic( TComPicYuv* pcUsPic, TComPicYuv* pcBasePic, TComPicYuv* pcTempPic )
{
  assert ( NTAPS_US_LUMA == 8 );
  assert ( NTAPS_US_CHROMA == 4 );

  Int i, j;

  for ( i = 0; i < 16; i++)
  {
    for ( j = 0; j < NTAPS_US_LUMA; j++)
    {
      m_iFilter16_luma[i][j]= m_iFilter16_luma_fixed[i][j];
    }

    for ( j = 0; j < NTAPS_US_CHROMA; j++)
    {
      m_iFilter16_chroma[ i][j] =  m_iFilter16_chroma_fixed[i][j];
    }
  }
  
 //========== Y component upsampling ===========

#if PIC_UPSAMPLE_WITH_CROPPING
  Int iBWidth   = pcBasePic->getWidth () - pcBasePic->getPicCropLeftOffset() - pcBasePic->getPicCropRightOffset();
  Int iBHeight  = pcBasePic->getHeight() - pcBasePic->getPicCropTopOffset() - pcBasePic->getPicCropBottomOffset();
  Int iBStride  = pcBasePic->getStride();

  Int iEWidth   = pcUsPic->getWidth () - pcUsPic->getPicCropLeftOffset() - pcUsPic->getPicCropRightOffset();
  Int iEHeight  = pcUsPic->getHeight() - pcUsPic->getPicCropTopOffset() - pcUsPic->getPicCropBottomOffset();
  Int iEStride  = pcUsPic->getStride();
#else
  Int iBWidth   = pcBasePic->getWidth ();
  Int iBHeight  = pcBasePic->getHeight();
  Int iBStride  = pcBasePic->getStride();

  Int iEWidth   = pcUsPic->getWidth ();
  Int iEHeight  = pcUsPic->getHeight();
  Int iEStride  = pcUsPic->getStride();
#endif

  Int   iShiftX = 16;
  Int   iShiftY = 16;

  Int   iPhaseX = 0;
  Int   iPhaseY = 0;

  Int   iAddX       = ( ( ( iBWidth * iPhaseX ) << ( iShiftX - 2 ) ) + ( iEWidth >> 1 ) ) / iEWidth + ( 1 << ( iShiftX - 5 ) );
  Int   iAddY       = ( ( ( iBHeight * iPhaseY ) << ( iShiftY - 2 ) ) + ( iEHeight >> 1 ) ) / iEHeight+ ( 1 << ( iShiftY - 5 ) );

  Int   iDeltaX     = 4 * iPhaseX;
  Int   iDeltaY     = 4 * iPhaseY;  


  Int iShiftXM4 = iShiftX - 4;
  Int iShiftYM4 = iShiftY - 4;

  Int   iScaleX     = ( ( iBWidth << iShiftX ) + ( iEWidth >> 1 ) ) / iEWidth;
  Int   iScaleY     = ( ( iBHeight << iShiftY ) + ( iEHeight >> 1 ) ) / iEHeight;

  Pel* piTempBufY = pcTempPic->getLumaAddr();
  Pel* piSrcBufY  = pcBasePic->getLumaAddr();
  Pel* piDstBufY  = pcUsPic->getLumaAddr();

  Pel* piSrcY;
  Pel* piDstY;

  assert ( iEWidth == 2*iBWidth || 2*iEWidth == 3*iBWidth );
  assert ( iEHeight == 2*iBHeight || 2*iEHeight == 3*iBHeight );

  pcBasePic->setBorderExtension(false);
  pcBasePic->extendPicBorder   (); // extend the border.

  Double sFactor = iEWidth * 1.0 / iBWidth;

  //========== horizontal upsampling ===========
  for( j = 0; j < iBHeight ; j++ )
  {
    piSrcY = &piSrcBufY[j*iBStride];
    piDstY = &piTempBufY[j*iEStride];
    for( i = 0; i < iEWidth; i++ )
    {
      Int refPos12 = (Int) ( i * 12 / sFactor );
      Int iRefPos = refPos12 / 12;
      Int iPhase = (refPos12 + 12) % 12; 

      Pel* pel = piSrcY + iRefPos -((NTAPS_US_LUMA>>1) - 1);
      Int* coeff = m_iFilter16_luma[iPhase];

      piDstY[i] = sum_luma_hor(pel, coeff);
    }
  }

  //========== vertical upsampling ===========
  pcTempPic->setBorderExtension(false);
  pcTempPic->setHeight(iBHeight);
  pcTempPic->extendPicBorder   (); // extend the border.
  pcTempPic->setHeight(iEHeight);

  Int iOffset = 1 << (US_FILTER_PREC*2 - 1); 

  for( i = 0; i < iEWidth; i++ )
  {
    piSrcY = &piTempBufY[i];
    piDstY = &piDstBufY[i];
    for( j = 0; j < iEHeight; j++ )
    {
      Int refPos12 = (Int) (j * 12 / sFactor);
      Int iRefPos = refPos12 / 12;
      Int iPhase = (refPos12 + 12) % 12;

      Pel* pel = piSrcY + (iRefPos -((NTAPS_US_LUMA>>1) - 1))*iEStride;
      Int* coeff = m_iFilter16_luma[iPhase];

      piDstY[j*iEStride] = Clip( (sum_luma_ver(pel, coeff, iEStride) + iOffset) >> (US_FILTER_PREC*2));
    }
  }


  //========== UV component upsampling ===========
  Pel* piTempBufU = pcTempPic->getCbAddr();
  Pel* piSrcBufU  = pcBasePic->getCbAddr();
  Pel* piDstBufU  = pcUsPic->getCbAddr();

  Pel* piTempBufV = pcTempPic->getCrAddr();
  Pel* piSrcBufV  = pcBasePic->getCrAddr();
  Pel* piDstBufV  = pcUsPic->getCrAddr();

  Pel* piSrcU;
  Pel* piDstU;
  Pel* piSrcV;
  Pel* piDstV;

  iEWidth  >>= 1;
  iEHeight >>= 1;

  iBWidth  >>= 1;
  iBHeight >>= 1;

  iBStride  = pcBasePic->getCStride();
  iEStride  = pcUsPic->getCStride();

  iShiftX = 16;
  iShiftY = 16;

  iPhaseX = 0;
  iPhaseY = 1;

  iAddX       = ( ( ( iBWidth * iPhaseX ) << ( iShiftX - 2 ) ) + ( iEWidth >> 1 ) ) / iEWidth + ( 1 << ( iShiftX - 5 ) );
  iAddY       = ( ( ( iBHeight * iPhaseY ) << ( iShiftY - 2 ) ) + ( iEHeight >> 1 ) ) / iEHeight+ ( 1 << ( iShiftY - 5 ) );

  iDeltaX     = 4 * iPhaseX;
  iDeltaY     = 4 * iPhaseY;


  iShiftXM4 = iShiftX - 4;
  iShiftYM4 = iShiftY - 4;

  iScaleX     = ( ( iBWidth << iShiftX ) + ( iEWidth >> 1 ) ) / iEWidth;
  iScaleY     = ( ( iBHeight << iShiftY ) + ( iEHeight >> 1 ) ) / iEHeight;

  //========== horizontal upsampling ===========

  for( j = 0; j < iBHeight ; j++ )
  {
    piSrcU = &piSrcBufU[j*iBStride];
    piDstU = &piTempBufU[j*iEStride];

    piSrcV = &piSrcBufV[j*iBStride];
    piDstV = &piTempBufV[j*iEStride];

    for( i = 0; i < iEWidth; i++ )
    {
      Int refPos12 = (Int) ( i * 12 / sFactor );
      Int iRefPos = refPos12 / 12;
      Int iPhase = (refPos12 + 12) % 12;

      Pel* pelU = piSrcU + iRefPos -((NTAPS_US_CHROMA>>1) - 1);
      Pel* pelV = piSrcV + iRefPos -((NTAPS_US_CHROMA>>1) - 1);
      Int* coeff = m_iFilter16_chroma[iPhase];

      piDstU[i] = sum_chroma_hor(pelU, coeff);
      piDstV[i] = sum_chroma_hor(pelV, coeff);
    }
  }

  //========== vertical upsampling ===========
  pcTempPic->setBorderExtension(false);
  pcTempPic->setHeight(iBHeight << 1);
  pcTempPic->extendPicBorder   (); // extend the border.
  pcTempPic->setHeight(iEHeight << 1);

  for( i = 0; i < iEWidth; i++ )
  {
    piSrcU = &piTempBufU[i];
    piDstU = &piDstBufU[i];

    piSrcV = &piTempBufV[i];
    piDstV = &piDstBufV[i];

    for( j = 0; j < iEHeight; j++ )
    {
      Int refPos12 = (Int) (j * 12 / sFactor);
      Int iRefPos = refPos12 / 12;
      Int iPhase = (refPos12 + 12) % 12;

      Pel* pelU = piSrcU + (iRefPos -((NTAPS_US_CHROMA>>1) - 1))*iEStride;
      Pel* pelV = piSrcV + (iRefPos -((NTAPS_US_CHROMA>>1) - 1))*iEStride;
      Int* coeff = m_iFilter16_chroma[iPhase];

      piDstU[j*iEStride] = Clip( (sum_chroma_ver(pelU, coeff, iEStride) + iOffset) >> (US_FILTER_PREC*2));
      piDstV[j*iEStride] = Clip( (sum_chroma_ver(pelV, coeff, iEStride) + iOffset) >> (US_FILTER_PREC*2));
    }
  }

  pcUsPic->setBorderExtension(false);
  pcUsPic->extendPicBorder   (); // extend the border.

  //Reset the Border extension flag
  pcUsPic->setBorderExtension(false);
  pcTempPic->setBorderExtension(false);
  pcBasePic->setBorderExtension(false);
}
#endif //SVC_EXTENSION
